'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
# XES TAGS
TAG_BOOLEAN = 'boolean'
TAG_CLASSIFIER = 'classifier'
TAG_DATE = 'date'
TAG_EVENT = 'event'
TAG_EXTENSION = 'extension'
TAG_FLOAT = 'float'
TAG_GLOBAL = 'global'
TAG_ID = 'id'
TAG_INT = 'int'
TAG_LIST = 'list'
TAG_LOG = 'log'
TAG_STRING = 'string'
TAG_TRACE = 'trace'
TAG_VALUES = 'values'

# XES 2.0 tags
TAG_CONTAINER = 'container'

# XES/INTERNAL KEYS
KEY_CHILDREN = 'children'
KEY_KEY = 'key'
KEY_KEYS = 'keys'
KEY_NAME = 'name'
KEY_PREFIX = 'prefix'
KEY_SCOPE = 'scope'
KEY_URI = 'uri'
KEY_VALUE = 'value'

DEFAULT_NAME_KEY = 'concept:name'
DEFAULT_INSTANCE_KEY = 'concept:instance'
DEFAULT_TIMESTAMP_KEY = 'time:timestamp'
DEFAULT_START_TIMESTAMP_KEY = 'start_timestamp'
DEFAULT_TRACEID_KEY = 'concept:name'
DEFAULT_RESOURCE_KEY = 'org:resource'
DEFAULT_TRANSITION_KEY = "lifecycle:transition"
DEFAULT_GROUP_KEY = 'org:group'
